# coding=utf-8
"""
tolerance decorator module
"""
__author__ = 'Alisue <lambdalisue@hashnote.net>'
import sys
from tolerance.utils import argument_switch_generator
from tolerance.functional import wraps


DEFAULT_TOLERATE_SWITCH = argument_switch_generator('fail_silently')
"""Default tolerate switch function"""


def tolerate(substitute=None, exceptions=None,
             switch=DEFAULT_TOLERATE_SWITCH):
    """
    A function decorator which makes a function fail silently

    To disable fail silently in a decorated function, specify
    ``fail_silently=False``.
    To disable fail silenlty in decorated functions globally, specify
    ``tolerate.disabled``.

    Parameters
    ----------
    fn : function
        A function which will be decorated.
    substitute : function or returning value
        A function used instead of :attr:`fn` or returning value
        when :attr:`fn` failed.
    exceptions : list of exceptions or None
        A list of exception classes or None. 
        If exceptions is specified, ignore exceptions only listed in this
        parameter and raise exception if the exception is not listed.
    switch : string, list/tuple, dict, function or None
        A switch function which determine whether silent the function failar.
        The function receive ``*args`` and ``**kwargs`` which will specified
        to :attr:`fn` and should return status (bool), args, and kwargs.
        If the function return ``False`` then agggressive decorated function
        worked as normal function (raise exception when there is exception).
        Default switch function is generated by
        :func:`argument_switch_generator` with
        ``argument_switch_generator('fail_silently')``
        so if ``fail_silently=False`` is specified to the function, the function
        works as noramlly.  
        
        **From Version 0.1.1**, when switch is specified as non functional value,
        :func:`argument_switch_generator` will be called with switch as
        arguments.
        If string is specified, the switch generator will be called as
        ``argument_switch_generator(switch)``.
        If list or tuple is specified, the switch generator will be called as
        ``argument_switch_generator(*switch)``.
        If dict is specified, the switch generator will be called as
        ``argument_switch_generator(**switch)``.


    Returns
    -------
    function
        A decorated function
        
    Examples
    --------
    >>> #
    >>> # use tolerate as a function wrapper
    >>> #
    >>> parse_int = tolerate()(int)
    >>> parse_int(0)
    0
    >>> parse_int("0")
    0
    >>> parse_int("zero") is None
    True
    >>> #
    >>> # use tolerate as a function decorator (PIP-318)
    >>> #
    >>> @tolerate(lambda x: x)
    ... def prefer_int(x):
    ...     return int(x)
    >>> prefer_int(0)
    0
    >>> prefer_int("0")
    0
    >>> prefer_int("zero")
    'zero'
    >>> #
    >>> # filter exceptions be ignored
    >>> #
    >>> @tolerate(exceptions=(KeyError, ValueError))
    ... def force_int(x):
    ...     string_numbers = {
    ...         'zero': 0,
    ...         'one': 1,
    ...         'two': 2,
    ...         'three': 3,
    ...         'four': 4,
    ...         'five': 5,
    ...         'six': 6,
    ...         'seven': 7,
    ...         'eight': 8,
    ...         'nine': 9
    ...     }
    ...     if isinstance(x, (int, float)):
    ...         return int(x)
    ...     elif isinstance(x, str):
    ...         if x in string_numbers:
    ...             return string_numbers[x]
    ...         elif x in ('ten', 'hundred', 'thousand'):
    ...             raise KeyError
    ...         raise ValueError
    ...     else:
    ...         raise AttributeError
    >>> force_int('zero')
    0
    >>> force_int('ten') is None    # KeyError
    True
    >>> force_int('foo') is None    # ValueError
    True
    >>> force_int(object)           # AttributeError
    Traceback (most recent call last):
        ...
    AttributeError
    >>> #
    >>> # disable tolerance by passing `fail_silently=False`
    >>> #
    >>> force_int('ten', fail_silently=False)   # KeyError
    Traceback (most recent call last):
        ...
    KeyError
    >>> #
    >>> # disable tolerance globally by setting `tolerate.disabled=True`
    >>> #
    >>> tolerate.disabled = True
    >>> force_int('foo')    # ValueError
    Traceback (most recent call last):
        ...
    ValueError
    >>> tolerate.disabled = False   # rollback
    >>> #
    >>> # Features from Version 0.1.1
    >>> # 
    >>> # specify switch as a string
    >>> parse_int_string = tolerate(switch='patient')(int)
    >>> parse_int_string('zero') is None
    True
    >>> parse_int_string('zero', patient=False)
    Traceback (most recent call last):
        ...
    ValueError: ...
    >>> # specify switch as a list
    >>> parse_int_list = tolerate(switch=['fail_silently', False])(int)
    >>> parse_int_list('zero')
    Traceback (most recent call last):
        ...
    ValueError: ...
    >>> parse_int_string('zero', fail_silently=True) is None
    True
    >>> # specify switch as a dict
    >>> parse_int_dict = tolerate(switch={'argument_name': 'aggressive',
    ...                                   'reverse': True})(int)
    >>> parse_int_dict('zero') is None
    True
    >>> parse_int_dict('zero', aggressive=False) is None
    True
    >>> parse_int_dict('zero', aggressive=True) is None
    Traceback (most recent call last):
        ...
    ValueError: ...
    """
    if switch:
        # create argument switch if switch is string or list or dict
        if isinstance(switch, basestring):
            switch = argument_switch_generator(switch)
        elif isinstance(switch, (list, tuple)):
            switch = argument_switch_generator(*switch)
        elif isinstance(switch, dict):
            switch = argument_switch_generator(**switch)
    # callable alternative because callable is removed in python 3
    is_callable = lambda x: hasattr(x, '__call__')
    def decorator(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            if getattr(tolerate, 'disabled', False):
                # the function has disabled so call normally.
                return fn(*args, **kwargs)
            if switch is not None:
                status, args, kwargs = switch(*args, **kwargs)
                if not status:
                    # the switch function return `False` so call noramlly.
                    return fn(*args, **kwargs)
            try:
                return fn(*args, **kwargs)
            except:
                e = sys.exc_info()[1]
                if exceptions is None or e.__class__ in exceptions:
                    if is_callable(substitute):
                        return substitute(*args, **kwargs)
                    return substitute
                raise e
        return inner
    return decorator

if __name__ == '__main__':
    import doctest; doctest.testmod()



# coding=utf-8
"""
tolerance

tolerance is a function decorator to make a tolerant function; a function
which does not raise any exceptions even there are exceptions.
This concept is quite useful for making stable product or ``prefer_int`` types
of code described in Usage section.

"""
__author__ = 'Alisue <lambdalisue@hashnote.net>'
__all__ = ('__version__', 'tolerate', 'argument_switch_generator')
import os.path
from pkg_resources import get_distribution
from pkg_resources import DistributionNotFound

DISTRIBUTION_NAME = 'tolerance'

# get version information from setup.py
try: 
    _dist = get_distribution(DISTRIBUTION_NAME)
    if not __file__.startswith(os.path.join(_dist.location,
                                            DISTRIBUTION_NAME)):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'Please install this project with setup.py'
else:
    __version__ = _dist.version

# shortcut
from tolerance.decorators import tolerate
from tolerance.utils import argument_switch_generator

VERSION = tuple(map(tolerate(lambda x: x)(int), __version__.split('.')))



# coding=utf-8
"""
tolerance utility module
"""
__author__ = 'Alisue <lambdalisue@hashnote.net>'

DEFAULT_ARGUMENT_NAME = 'fail_silently'

def argument_switch_generator(argument_name = None,
                              default=True, reverse=False,
                              keep=False):
    """
    Create switch function which return the status from specified named argument

    Parameters
    ----------
    argument_name : string or None
        An argument name which is used to judge the status.
        If ``None`` is specified, the value of
        ``tolerance.utils.DEFAULT_ARGUMENT_NAME`` will be used instead.
    default : boolean
        A default value of this switch function.
        It is used when specifid ``**kwargs`` does not have named argument
    reverse : boolean
        Reverse the status (Default: ``False``)
    keep : boolean
        If it is ``True``, keep named argument in ``**kwargs``.

    Returns
    -------
    function
        A switch function which return status, args, and kwargs respectively.

    Examples
    --------
    >>> #
    >>> # generate switch function with default parameters
    >>> #
    >>> fn = argument_switch_generator('fail_silently')
    >>> # return `default` value and specified *args and **kwargs when
    >>> # `fail_silently` is not specified in **kwargs
    >>> fn() == (True, tuple(), {})
    True
    >>> # return `fail_silently` value when it is specified
    >>> fn(fail_silently=True) == (True, tuple(), {})
    True
    >>> fn(fail_silently=False) == (False, tuple(), {})
    True
    >>> #
    >>> # generate switch function with `default=False`
    >>> #
    >>> fn = argument_switch_generator('fail_silently', default=False)
    >>> # return `default` value so `False` is returned back
    >>> fn() == (False, tuple(), {})
    True
    >>> #
    >>> # generate switch function with `reverse=True`
    >>> #
    >>> fn = argument_switch_generator('fail_silently', reverse=True)
    >>> # `default` value is independent from `reverse=True`
    >>> fn() == (True, tuple(), {})
    True
    >>> # `fail_silently` value is influenced by `reverse=True`
    >>> fn(fail_silently=True) == (False, tuple(), {})
    True
    >>> fn(fail_silently=False) == (True, tuple(), {})
    True
    >>> #
    >>> # generate switch function with `keep=True`
    >>> #
    >>> fn = argument_switch_generator('fail_silently', keep=True)
    >>> # `fail_silently` attribute remains even in returned back kwargs
    >>> status, args, kwargs = fn(fail_silently=True)
    >>> 'fail_silently' in kwargs
    True
    """
    if argument_name is None:
        argument_name = DEFAULT_ARGUMENT_NAME
    def switch_function(*args, **kwargs):
        if argument_name in kwargs:
            if keep:
                status = kwargs.get(argument_name)
            else:
                status = kwargs.pop(argument_name)
            if reverse:
                status = not status
        else:
            status = default
        return bool(status), args, kwargs
    return switch_function


if __name__ == '__main__':
    import doctest; doctest.testmod()



try:
    # functools was introduced from Python 2.5
    from functools import wraps
except ImportError:
    # License for code in this file that was taken from Python 2.5.

    # PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
    # --------------------------------------------
    #
    # 1. This LICENSE AGREEMENT is between the Python Software Foundation
    # ("PSF"), and the Individual or Organization ("Licensee") accessing and
    # otherwise using this software ("Python") in source or binary form and
    # its associated documentation.
    #
    # 2. Subject to the terms and conditions of this License Agreement, PSF
    # hereby grants Licensee a nonexclusive, royalty-free, world-wide
    # license to reproduce, analyze, test, perform and/or display publicly,
    # prepare derivative works, distribute, and otherwise use Python
    # alone or in any derivative version, provided, however, that PSF's
    # License Agreement and PSF's notice of copyright, i.e., "Copyright (c)
    # 2001, 2002, 2003, 2004, 2005, 2006, 2007 Python Software Foundation;
    # All Rights Reserved" are retained in Python alone or in any derivative
    # version prepared by Licensee.
    #
    # 3. In the event Licensee prepares a derivative work that is based on
    # or incorporates Python or any part thereof, and wants to make
    # the derivative work available to others as provided herein, then
    # Licensee hereby agrees to include in any such work a brief summary of
    # the changes made to Python.
    #
    # 4. PSF is making Python available to Licensee on an "AS IS"
    # basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
    # IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND
    # DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
    # FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT
    # INFRINGE ANY THIRD PARTY RIGHTS.
    #
    # 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
    # FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS
    # A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON,
    # OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.
    #
    # 6. This License Agreement will automatically terminate upon a material
    # breach of its terms and conditions.
    #
    # 7. Nothing in this License Agreement shall be deemed to create any
    # relationship of agency, partnership, or joint venture between PSF and
    # Licensee.  This License Agreement does not grant permission to use PSF
    # trademarks or trade name in a trademark sense to endorse or promote
    # products or services of Licensee, or any third party.
    #
    # 8. By copying, installing or otherwise using Python, Licensee
    # agrees to be bound by the terms and conditions of this License
    # Agreement.


    def curry(_curried_func, *args, **kwargs):
        def _curried(*moreargs, **morekwargs):
            return _curried_func(*(args+moreargs), **dict(kwargs, **morekwargs))
        return _curried

    ### Begin from Python 2.5 functools.py ########################################

    # Summary of changes made to the Python 2.5 code below:
    #   * swapped ``partial`` for ``curry`` to maintain backwards-compatibility
    #     in Django.

    # Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007 Python Software Foundation.
    # All Rights Reserved.

    ###############################################################################

    # update_wrapper() and wraps() are tools to help write
    # wrapper functions that can handle naive introspection

    WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__doc__')
    WRAPPER_UPDATES = ('__dict__',)
    def update_wrapper(wrapper,
                    wrapped,
                    assigned = WRAPPER_ASSIGNMENTS,
                    updated = WRAPPER_UPDATES):
        """Update a wrapper function to look like the wrapped function

        wrapper is the function to be updated
        wrapped is the original function
        assigned is a tuple naming the attributes assigned directly
        from the wrapped function to the wrapper function (defaults to
        functools.WRAPPER_ASSIGNMENTS)
        updated is a tuple naming the attributes off the wrapper that
        are updated with the corresponding attribute from the wrapped
        function (defaults to functools.WRAPPER_UPDATES)
        """
        for attr in assigned:
            setattr(wrapper, attr, getattr(wrapped, attr))
        for attr in updated:
            getattr(wrapper, attr).update(getattr(wrapped, attr))
        # Return the wrapper so this can be used as a decorator via curry()
        return wrapper

    def wraps(wrapped,
            assigned = WRAPPER_ASSIGNMENTS,
            updated = WRAPPER_UPDATES):
        """Decorator factory to apply update_wrapper() to a wrapper function

        Returns a decorator that invokes update_wrapper() with the decorated
        function as the wrapper argument and the arguments to wraps() as the
        remaining arguments. Default arguments are as for update_wrapper().
        This is a convenience function to simplify applying curry() to
        update_wrapper().
        """
        return curry(update_wrapper, wrapped=wrapped,
                    assigned=assigned, updated=updated)
